from pydantic.dataclasses import dataclass as pd_dataclass

from captur_ml.predict import PredictionJobConfig, LivePredictionJob, BatchPredictionJob, LivePredictionJobConfig, BatchPredictionJobConfig
from captur_ml.core.dtypes import ImageOrImageList


@pd_dataclass
class CapturMLConfig:
    """
    Configuration for the CapturML client.
    """
    pass


class CapturML:
    """
    The CapturML class is the main entry point for the CapturML machine learning library. It unifies the APIs for the different machine learning tasks (training, evaluation, prediction) and provides a single interface to interact with.
    """
    def __init__(self, config: CapturMLConfig | None = None):
        self.config = config

    def predict(self, images: ImageOrImageList, config: PredictionJobConfig):
        """Generates predictions for an image or a set of images.

        Args:
            images (ImageOrImageList): The images to predict on.
            config (PredictionJobConfig): A configuration object for the prediction job. The choice of configuration object depends on the type of prediction job and will determine the type of prediction job.

        Raises:
            ValueError: If the prediction job config is not of a known type.

        Returns:
            Any: The result of the prediction job. If the prediction job is a live prediction job, the result will be a list of predictions. If the prediction job is a batch prediction job, the result will be a promise type object that can be used to retrieve the result of the prediction job.
        """
        match config:
            case LivePredictionJobConfig():
                job = LivePredictionJob(config)
            case BatchPredictionJobConfig():
                job = BatchPredictionJob(config)
            case other:
                raise ValueError(f"Unknown prediction job config type: {other}. Supported types are: 'LivePredictionJobConfig', 'BatchPredictionJobConfig'")
        
        return job.start(images)

    def train(self, images: ImageOrImageList, config):
        """This is a stub method for training a model.

        Args:
            images (ImageOrImageList): The images to train on.
            config (TrainingJobConfig): A configuration object for the training job.

        Raises:
            NotImplementedError
        """
        raise NotImplementedError()
    
    def evaluate(self, images: ImageOrImageList, config):
        """This is a stub method for evaluating a model.
            
            Args:
                images (ImageOrImageList): The images to evaluate on.
                config (EvaluationJobConfig): A configuration object for the evaluation job.
    
            Raises:
                NotImplementedError
            """
        raise NotImplementedError()



from enum import Enum
from pydantic.dataclasses import dataclass as pd_dataclass

from captur_ml.core.dtypes import ImageOrImageList
from captur_ml.core.entities.endpoint import Endpoint
from captur_ml.core.services.aiplatform.live_predict import (
    get_image_classification_prediction_from_deployed_automl,
    convert_automl_to_captur,
)
from captur_ml.core.dtypes.result import ImageClassificationResult


class PredictionService(Enum):
    AUTOML = "automl"


@pd_dataclass
class LivePredictionJobConfig:
    """
    Configuration for a live prediction job.
    """
    #: The endpoint to use for the prediction.
    endpoint: Endpoint


class LivePredictionJob:
    """
    A live prediction job is a job that runs on an image or set of images and returns the results of the prediction job. It is first initialized with a configuration object and then started with the chosen image(s).
    """
    def __init__(self, config: LivePredictionJobConfig):
        self.config = config
        self.service: PredictionService = PredictionService.AUTOML

    def start(self, images: ImageOrImageList) -> list[ImageClassificationResult]:
        """Starts the (sync) live prediction job.

        Args:
            images (ImageOrImageList): The images to predict on.

        Returns:
            list[ImageClassificationResult]: The result of the prediction job.
        """
        match self.service:
            case PredictionService.AUTOML:
                result = get_image_classification_prediction_from_deployed_automl(
                    images=images,
                    endpoint=self.config.endpoint,
                )
                result = convert_automl_to_captur(images, result)

                return result



import abc
import math
import torch
from sklearn.metrics import (
    accuracy_score,
    confusion_matrix,
    f1_score,
    precision_score,
    recall_score,
)
from torchmetrics import (
    Accuracy,
    F1Score,
    Precision,
    Recall,
    ConfusionMatrix,
)

from typing import Dict, List, Union


class Metrics(metaclass=abc.ABCMeta):
    def __init__(
        self, labels, predictions, metrics: List[str] = None, is_external: bool = False
    ) -> None:
        super().__init__()
        self.labels = labels
        self.predictions = predictions
        self.metrics = metrics
        self.is_external = is_external

    @abc.abstractmethod
    def external_metrics(self) -> Dict:
        pass

    @abc.abstractmethod
    def internal_metrics(self) -> Dict:
        pass

    def calculate(self) -> Dict:
        metric_group = self.external_metrics()
        if not self.is_external:
            metric_group = self.internal_metrics()

        metric_list = metric_group.keys()
        if self.metrics is not None and not self.is_external:
            metric_list = self.metrics

        metrics_dict = {}
        for metric in metric_list:
            metrics_dict[metric] = metric_group[metric]()
        return metrics_dict

    @staticmethod
    def dictify_per_class_metrics(classes, result) -> Dict[str, Union[float, None]]:
        return {class_name: score for class_name, score in zip(classes, result)}

    @staticmethod
    def csvify_confusion_matrix(
        classes: List[str], confusion_matrix: List[List[int]]
    ) -> List[str]:
        csv = []
        num_classes = len(classes)
        for i in range(num_classes):
            csv_string = str(classes[i]) + ","
            for j in range(num_classes):
                csv_string += str(confusion_matrix[i][j])
                if j < num_classes - 1:
                    csv_string += ","
            csv.append(csv_string)
        return csv

    @staticmethod
    def noneify_nans(result_list: List) -> List:
        for i in range(len(result_list)):
            if math.isnan(result_list[i]):
                result_list[i] = None
        return result_list


class ClassificationMetrics(Metrics):
    def __init__(
        self,
        predictions: torch.Tensor,
        labels: torch.Tensor,
        classes: List[str],
        metrics: List[str] = None,
        top_k: int = None,
        is_external: bool = True,
    ) -> None:
        """Calculates classification metrics from predictions and labels.

        Args:
            predictions (torch.Tensor): Array of model predictions of shape (N, C) where N is
                                        the number of examples and C is the number of classes.
            labels (torch.Tensor): Array of ground truth labels of shape (N), where N is the
                                   number of examples.
            classes (List[str]): Array of class names.
            metrics (List[str], optional): List of metrics to be calculated. Defaults to None.
            top_k (int, optional): Number of the highest probability score predictions considered
                                   for choosing the correct label. Defaults to None.
            is_external (bool, optional): Whether to return only those metrics suitable for external
                                          purposes. Defaults to False.

        Raises:
            TypeError: Raised if the predictions or labels are not torch.Tensor.
            ValueError: Raised if the value for top_k is less than 1.
            ValueError: Raised if the value for top_k is greater or equal to the number of classes in the predictions.
        """
        super().__init__(labels, predictions, metrics, is_external)

        if not isinstance(predictions, torch.Tensor):
            raise TypeError(
                f"Invalid type for predictions {type(predictions)}, must be type torch.Tensor"
            )
        if not isinstance(labels, torch.Tensor):
            raise TypeError(
                f"Invalid type for labels {type(labels)}, must be type torch.Tensor"
            )

        self.classes = classes
        self.num_classes = len(classes)

        if top_k is None:
            top_k = 1
        if top_k <= 0:
            raise ValueError(f"Invalid value for top K {top_k}")
        self.top_k = top_k

        if self.top_k >= self.num_classes:
            raise ValueError(
                f"top_k ({self.top_k}) must be less than number of classes ({self.num_classes})"
            )

        self.is_external = is_external

    def accuracy(self) -> float:
        """Calculates top-k accuracy on the predictions and labels.

        Returns:
            float: Top-k accuracy.
        """
        accuracy_fn = Accuracy(num_classes=self.num_classes, top_k=self.top_k)
        return accuracy_fn(self.predictions, self.labels).item()

    def conf_matrix(self) -> List[str]:
        """Calculates confusion matrix for the predictions and labels.

        Returns:
            List: Confusion matrix.
        """
        confusion_matrix_fn = ConfusionMatrix(num_classes=self.num_classes)
        conf_mat = confusion_matrix_fn(self.predictions, self.labels).tolist()
        return self.csvify_confusion_matrix(self.classes, conf_mat)

    def f1_score(self) -> float:
        """Calculates top-k f1 score on the predictions and labels.

        Returns:
            float: Top-k f1 score.
        """
        f1_score_fn = F1Score(num_classes=self.num_classes, top_k=self.top_k)
        return f1_score_fn(self.predictions, self.labels).item()

    def f1_score_per_class(self) -> Dict[str, Union[float, None]]:
        """Calculates top-k f1 score per class on the predictions and labels.

        Returns:
            Dict[str, float]: Dictionary of top-k f1 score per class.
        """
        f1_score_per_class_fn = F1Score(
            num_classes=self.num_classes, top_k=self.top_k, average=None
        )
        result = f1_score_per_class_fn(self.predictions, self.labels).tolist()
        result = self.noneify_nans(result)
        return self.dictify_per_class_metrics(self.classes, result)

    def precision(self) -> float:
        """Calculates top-k precision on the predictions and labels.

        Returns:
            float: Top-k precision.
        """
        self.precision = Precision(num_classes=self.num_classes, top_k=self.top_k)
        return self.precision(self.predictions, self.labels).item()

    def precision_per_class(self) -> Dict[str, Union[float, None]]:
        """Calculates top-k precision per class on the predictions and labels.

        Returns:
            Dict[str, float]: Dictionary of top-k precision per class.
        """
        self.precision_per_class = Precision(
            num_classes=self.num_classes, top_k=self.top_k, average=None
        )
        result = self.precision_per_class(self.predictions, self.labels).tolist()
        result = self.noneify_nans(result)
        return self.dictify_per_class_metrics(self.classes, result)

    def recall(self) -> float:
        """Calculates top-k recall on the predictions and labels.

        Returns:
            float: Top-k recall.
        """
        self.recall = Recall(num_classes=self.num_classes, top_k=self.top_k)
        return self.recall(self.predictions, self.labels).item()

    def recall_per_class(self) -> Dict[str, Union[float, None]]:
        """Calculates top-k recall per class on the predictions and labels.

        Returns:
            Dict[str, float]: Dictionary of top-k recall per class.
        """
        self.recall_per_class = Recall(
            num_classes=self.num_classes, stop_k=self.top_k, average=None
        )
        result = self.recall_per_class(self.predictions, self.labels).tolist()
        result = self.noneify_nans(result)
        return self.dictify_per_class_metrics(self.classes, result)

    def external_metrics(self) -> Dict:
        return {"accuracy": self.accuracy}

    def internal_metrics(self) -> Dict:
        return {
            "accuracy": self.accuracy,
            "f1_score": self.f1_score,
            "f1_score_per_class": self.f1_score_per_class,
            "precision": self.precision,
            "precision_per_class": self.precision_per_class,
            "recall": self.recall,
            "recall_per_class": self.recall_per_class,
            "confusion_matrix": self.conf_matrix,
        }


class AuditMetrics(Metrics):
    def __init__(
        self,
        audit_labels: List[List[str]],
        predictions: List[str],
        is_external: bool = True,
    ) -> None:
        super().__init__(
            labels=audit_labels, predictions=predictions, is_external=is_external
        )

    def audit_accuracy(self) -> float:
        count_correct = 0
        for al, p in zip(self.labels, self.predictions):
            count_correct += p in al
        return count_correct / len(self.predictions)

    def audit_precision_per_class(self) -> Dict[str, Union[float, None]]:
        all_predicted_classes = list(set(self.predictions))
        all_label_classes = list(
            set([num for sublist in self.labels for num in sublist])
        )
        precision_dict = {}
        for cls in all_predicted_classes:
            audit = 0
            count = 0
            for i in range(len(self.predictions)):
                if self.predictions[i] == cls:
                    count += 1
                    if self.predictions[i] in self.labels[i]:
                        audit += 1
            precision = audit / count
            precision_dict[cls] = precision
        for lbl in all_label_classes:
            if lbl not in precision_dict:
                precision_dict[lbl] = None
        return precision_dict

    def external_metrics(self) -> Dict:
        return {"accuracy": self.audit_accuracy}

    def internal_metrics(self) -> Dict:
        return {
            "accuracy": self.audit_accuracy,
            "precision_per_class": self.audit_precision_per_class,
        }


class MappedClassificationMetrics(Metrics):
    def __init__(
        self,
        labels: List[str],
        predictions: List[str],
        classes: List[str],
        metrics: List[str] = None,
        is_external: bool = True,
    ) -> None:
        super().__init__(labels, predictions, metrics, is_external)
        self.classes = classes

    def accuracy(self) -> float:
        return accuracy_score(
            self.labels,
            self.predictions,
        )

    def confusion_matrix(self) -> List[str]:
        conf_mat = confusion_matrix(
            self.labels, self.predictions, labels=self.classes
        ).tolist()
        return self.csvify_confusion_matrix(self.classes, conf_mat)

    def f1_score(self) -> float:
        return f1_score(
            self.labels,
            self.predictions,
            average="micro",
            labels=self.classes,
            zero_division=0,
        )

    def f1_score_per_class(self) -> Dict[str, Union[float, None]]:
        result = f1_score(
            self.labels,
            self.predictions,
            average=None,
            labels=self.classes,
            zero_division=0,
        )
        return self.dictify_per_class_metrics(self.classes, result)

    def precision(self) -> float:
        return precision_score(
            self.labels,
            self.predictions,
            average="micro",
            labels=self.classes,
            zero_division=0,
        )

    def precision_per_class(self) -> Dict[str, Union[float, None]]:
        result = precision_score(
            self.labels,
            self.predictions,
            average=None,
            labels=self.classes,
            zero_division=0,
        )
        return self.dictify_per_class_metrics(self.classes, result)

    def recall(self) -> float:
        return recall_score(
            self.labels,
            self.predictions,
            average="micro",
            labels=self.classes,
            zero_division=0,
        )

    def recall_per_class(self) -> Dict[str, Union[float, None]]:
        result = recall_score(
            self.labels,
            self.predictions,
            average=None,
            labels=self.classes,
            zero_division=0,
        )
        return self.dictify_per_class_metrics(self.classes, result)

    def external_metrics(self) -> Dict:
        return {"accuracy": self.accuracy}

    def internal_metrics(self) -> Dict:
        return {
            "accuracy": self.accuracy,
            "f1_score": self.f1_score,
            "f1_score_per_class": self.f1_score_per_class,
            "precision": self.precision,
            "precision_per_class": self.precision_per_class,
            "recall": self.recall,
            "recall_per_class": self.recall_per_class,
            "confusion_matrix": self.confusion_matrix,
        }


class DetectionMetrics(Metrics):
    pass


class SemSegMetrics(Metrics):
    pass



from pydantic.dataclasses import dataclass as pd_dataclass

from .image import Image
from .label import ClassLabel
from .prediction import PredictionSet


@pd_dataclass
class TrainingData:
    #: The list of images used for training
    images: list[Image]
    #: The list of class labels used for training
    labels: list[ClassLabel]


@pd_dataclass
class PredictionData:
    #: The list of images used for prediction
    images: list[Image]


@pd_dataclass
class EvaluationData:
    #: The list of images used for evaluation
    images: list[Image]
    #: The list of class labels used for evaluation
    labels: list[ClassLabel]
    #: The list of classification predictions used for evaluation
    predictions: PredictionSet



from captur_ml.evaluate.dataset import (
    ClassificationDataset,
    MappedClassificationDataset,
)
from captur_ml.evaluate.metrics import (
    AuditMetrics,
    ClassificationMetrics,
    MappedClassificationMetrics,
)

from typing import Dict, List, Union


def evaluate_classification_dataset_standard(
    dataset: ClassificationDataset,
    metric_list: List[str] = None,
    is_external: bool = True,
    top_k: int = None,
) -> Dict[str, float]:

    metrics = ClassificationMetrics(
        predictions=dataset.torch_predictions,
        labels=dataset.torch_labels,
        classes=dataset.classes,
        metrics=metric_list,
        top_k=top_k,
        is_external=is_external,
    ).calculate()
    return metrics


def evaluate_classification_dataset_audit(
    dataset: ClassificationDataset,
    is_external: bool = True,
) -> Union[Dict[str, float], None]:

    metrics = None
    if dataset.audit_labels:
        metrics = AuditMetrics(
            audit_labels=dataset.audit_labels,
            predictions=dataset.top_predictions,
            is_external=is_external,
        ).calculate()
    return metrics


def evaluate_mapped_classification_dataset_standard(
    dataset: ClassificationDataset,
    mapping: Union[Dict[str, str], None] = None,
    metric_list: List[str] = None,
    is_external: bool = True,
) -> Dict[str, float]:

    metrics = None
    if mapping:
        mapped_dataset = MappedClassificationDataset.map_dataset(
            dataset=dataset,
            mapping=mapping,
        )
        metrics = MappedClassificationMetrics(
            labels=mapped_dataset.labels,
            predictions=mapped_dataset.predictions,
            classes=mapped_dataset.classes,
            metrics=metric_list,
            is_external=is_external,
        ).calculate()
    return metrics


def evaluate_mapped_classification_dataset_audit(
    dataset: ClassificationDataset,
    mapping: Union[Dict[str, str], None] = None,
    is_external: bool = True,
) -> Dict[str, float]:

    metrics = None
    if mapping and dataset.audit_labels:
        mapped_dataset = MappedClassificationDataset.map_dataset(
            dataset=dataset,
            mapping=mapping,
        )
        metrics = AuditMetrics(
            audit_labels=mapped_dataset.audit_labels,
            predictions=mapped_dataset.predictions,
            is_external=is_external,
        ).calculate()
    return metrics



from abc import ABC, abstractmethod
from pydantic.dataclasses import dataclass

from captur_ml.core.dtypes.generic import ObjectWithId


@dataclass
class BaseEndpoint(ABC, ObjectWithId):
    """A base class for endpoints. Endpoints are used to make predictions on images."""

    @property
    @abstractmethod
    def exists(self):
        pass

    @property
    @abstractmethod
    def is_active(self):
        pass

    @property
    @abstractmethod
    def models(self):
        pass



from typing import TypeAlias

from .image import Image
from .label import ClassLabel, AuditLabel, BoundingBoxLabel
from .prediction import ClassPrediction, PredictionSet

ImageOrImageList: TypeAlias = Image | list[Image]
Label: TypeAlias = ClassLabel | AuditLabel | BoundingBoxLabel

__all__ = [
    "AuditLabel",
    "ClassLabel",
    "ClassPrediction",
    "PredictionSet",
    "Image",
    "ImageOrImageList",
    "Label",
]



from pydantic.dataclasses import dataclass

from google.cloud.aiplatform import Endpoint
from google.cloud.aiplatform_v1beta1.types import DeployedModel

from google.api_core import exceptions as google_exceptions

from .base import BaseEndpoint
from captur_ml.core.exceptions import (
    GoogleCloudVertexAIEndpointDoesNotExistError,
    GoogleCloudVertexAIEndpointNoModelDeployedError,
)


@dataclass
class VertexAIEndpoint(BaseEndpoint):
    """An object that represents a Vertex AI endpoint. An endpoint is a deployed model that can be used for prediction."""

    #: The GCP project name.
    project: str
    #: The global location of the VertexAI Endpoint.
    location: str
    #: Run checks to see if the endpoint exists and is active on initialization.
    check_on_init: bool = False

    def __post_init__(self):
        self.endpoint = Endpoint(
            endpoint_name=self.id, project=self.project, location=self.location
        )

        if self.check_on_init:
            if not self.exists:
                raise GoogleCloudVertexAIEndpointDoesNotExistError()
            if not self.is_active:
                raise GoogleCloudVertexAIEndpointNoModelDeployedError()

    @property
    def exists(self) -> bool:
        """Checks if the VertexAI Endpoint exists."""
        try:
            self.endpoint.list_models()
        except (google_exceptions.NotFound, google_exceptions.InvalidArgument):
            return False
        return True

    @property
    def is_active(self) -> bool:
        """Checks if the VertexAI Endpoint exists and has at least one model serving requests."""
        if not self.exists:
            return False

        return len(self.models) > 0

    @property
    def models(self) -> list[DeployedModel]:
        """Returns a list of models deployed to the VertexAI Endpoint."""
        return self.endpoint.list_models()



# ============ PubSub Exceptions ============ #
class GoogleCloudPubSubTopicDoesNotExistError(Exception):
    pass


# ============ Google Cloud Storage Exceptions ============ #
class GoogleCloudStoragePermissionError(Exception):
    pass


class GoogleCloudStorageResourceNotFoundError(Exception):
    pass


class GoogleCloudStorageBucketNotFoundError(Exception):
    pass


# ============ Google Cloud Vertex AI Endpoint Exceptions ============ #
class GoogleCloudVertexAIEndpointDoesNotExistError(Exception):
    pass


class GoogleCloudVertexAIEndpointImageTooLargeError(Exception):
    pass


class GoogleCloudVertexAIEndpointNoModelDeployedError(Exception):
    pass


class GoogleCloudVertexAIEndpointCorruptedImageError(Exception):
    pass


# ============ Google Cloud Vertex AI General Exceptions ============ #
class GoogleCloudVertexAIModelDoesNotExistError(Exception):
    pass


class GoogleCloudVertexAIDatasetDoesNotExistError(Exception):
    pass


class GoogleCloudVertexAIResourceDoesNotExistError(Exception):
    pass


class GoogleCloudVertexAICompatibilityError(Exception):
    pass


class GoogleCloudVertexAIBatchGCSSourceDoesNotExistError(Exception):
    pass


# ============ Sentry Errors ============ #
class SentryDSNNotProvidedError(Exception):
    pass



from typing import Dict


def parse_batch_predict_request(batch_predict_request: Dict) -> Dict:
    """
    Parses the batch predict request sent to the ML Gateway into the form
    necessary to be received by the Batch Predict Cloud Function.
    Note: currently just identity function.

    Args:
        batch_predict_request (Dict): Dictionary version of the batch predict request sent to the MLGateway.

    Returns:
        Dict: Batch Predict request to be sent to the Batch Predict Cloud Function
    """
    return batch_predict_request


def parse_live_predict_request(live_predict_request: Dict) -> Dict:
    """
    Parses the live predict request sent to the ML Gateway into the form
    necessary to be received by the Live Predict Cloud Function.
    Note: currently just identity function.

    Args:
        live_predict_request (Dict): Dictionary version of the Live Predict request sent to the MLGateway.

    Returns:
        Dict: Live Predict request to be sent to the Live Predict Cloud Function
    """
    return live_predict_request


def parse_mongo_model_evaluate_request(model_evaluate_request: Dict) -> Dict:
    """
    Parses the model evaluate request sent to the ML Gateway into the form
    necessary to be received by the model evaluate Cloud Function.

    Args:
        model_evaluate_request (Dict): Dictionary version of the model evaluate request sent to the MLGateway.

    Returns:
        Dict: Model Evaluate request to be sent to the Model Evaluate Cloud Function
    """
    audit_labels = [
        image["audit_labels"] for image in model_evaluate_request["data"]["images"]
    ]
    if None in audit_labels:
        audit_labels = None
    parsed_request = {
        "labels": [
            image["label"] for image in model_evaluate_request["data"]["images"]
        ],
        "audit_labels": audit_labels,
        "predictions": [
            image["predictions"] for image in model_evaluate_request["data"]["images"]
        ],
        "webhooks": model_evaluate_request["meta"]["webhooks"],
        "request_id": model_evaluate_request["meta"]["request_id"],
        "metrics": model_evaluate_request["meta"]["metrics"],
        "is_external": model_evaluate_request["meta"]["is_external"],
    }
    return parsed_request


def parse_model_train_request(model_train_request: Dict) -> Dict:
    """
    Parses the model train request sent to the ML Gateway into the form
    necessary to be received by the Model Train Cloud Function.
    Note: currently just identity function.

    Args:
        model_train_request (Dict): Dictionary version of the model train request sent to the MLGateway.

    Returns:
        Dict: Model Train request to be sent to the Model Train Cloud Function
    """
    return model_train_request


def parse_model_deploy_request(model_deploy_request: Dict) -> Dict:
    """
    Parses the model deploy request sent to the ML Gateway into the form
    necessary to be received by the Model Deploya Cloud Function.
    Note: currently just identity function.

    Args:
        model_deploy_request (Dict): Dictionary version of the model deploy request sent to the MLGateway.

    Returns:
        Dict: Model Deploy request to be sent to the Model Deploy Cloud Function
    """
    return model_deploy_request



import numpy as np
import torch

from captur_ml.core.dtypes import (
    Image,
    ClassLabel,
    PredictionSet,
    AuditLabel,
)

# TODO: This section of code needs to be updated to use the new
#   pydantic dataclasses.


class ClassificationDataset(object):
    def __init__(
        self,
        labels: list[ClassLabel],
        prediction_set: PredictionSet,
        audit_labels: list[AuditLabel] = None,
    ):
        """Class for processing classification labels and predictions.

        Args:
            labels (list[ClassLabel]): list of ground truth labels for the images.
            prediction_set (PredictionSet): list of model predictions for the images.

        Raises:
            ValueError: Raised if the prediction or label array is empty.
            ValueError: Raised if the number of predictions does not match the number of labels.
            TypeError: Raised if:
                images is a list of type Image
                labels is not a list of type ClassLabel
                prediction_set is not a list of type PredictionSet
            ValueError: Raised if the number of classes in the labels is greater than the number of classes in the predictions.
        """
        if len(prediction_set) < 1:
            raise ValueError("Predictions cannot be an empty array")
        if len(labels) < 1:
            raise ValueError("Labels cannot be an empty array")
        if len(labels) != len(prediction_set):
            raise ValueError(
                f"Number of labels {len(labels)} does not match number of predictions {len(prediction_set)}"
            )

        if not isinstance(prediction_set, PredictionSet):
            raise TypeError(
                f"Prediction sets is type {type(prediction_set)} but must be of type PredictionSet (see captur_ml_sdk.dtypes)"
            )
        if not isinstance(labels[0], ClassLabel):
            raise TypeError(
                f"Labels is type {type(labels[0])} but must be of type ClassLabel (see captur_ml_sdk.dtypes)"
            )

        self.labels = [label.name for label in labels]

        if len(np.unique(self.labels)) > len(prediction_set.predictions):
            raise ValueError(
                "Number of classes in labels is greater than the number of classes in predictions"
            )
        self.audit_labels = None
        if audit_labels:
            self.audit_labels = [label.names for label in audit_labels]
        self.prediction_set = prediction_set
        self.classes = [pred.name for pred in self.prediction_set.predictions]

    @property
    def class_encoding(self):
        return {class_name: i for i, class_name in enumerate(self.classes)}

    @property
    def torch_labels(self):
        encoded_labels = [self.class_encoding[label] for label in self.labels]
        return torch.tensor(encoded_labels)

    @property
    def torch_predictions(self):
        ordered_prediction_sets = sorted(
            self.prediction_set.predictions,
            key=lambda pred: self.class_encoding[pred.name],
        )
        ordered_confidences = [
            pred.confidence for pred in self.prediction_set.predictions
        ]
        return torch.tensor(ordered_confidences)

    @property
    def top_predictions(self):
        return self.prediction_set.predictions[0].name


class MappedClassificationDataset(object):
    def __init__(
        self,
        labels: list[str],
        predictions: list[str],
        audit_labels: list[list[str]],
        classes: list[str],
    ):
        self.predictions = predictions
        self.labels = labels
        self.audit_labels = audit_labels
        self.classes = classes

    @classmethod
    def map_dataset(cls, dataset: ClassificationDataset, mapping: dict[str, str]):
        mapped_preds = [mapping[cls] for cls in dataset.top_predictions]
        mapped_labels = [mapping[cls] for cls in dataset.labels]
        mapped_audit_labels = None
        if dataset.audit_labels:
            mapped_audit_labels = cls.map_nested_array(dataset.audit_labels, mapping)
        mapped_classes = cls.get_unique_mapped_classes(mapping)
        return cls(mapped_preds, mapped_labels, mapped_audit_labels, mapped_classes)

    @staticmethod
    def map_nested_array(nested_array: list[list[str]], mapping: dict[str, str]):
        mapped_nested_array = []
        for array in nested_array:
            mapped_array = []
            for item in array:
                try:
                    mapped_array.append(mapping[item])
                except KeyError:
                    raise ValueError(f"{item} in audit labels not present in mapping")
            mapped_nested_array.append(mapped_array)
        return mapped_nested_array

    @staticmethod
    def get_unique_mapped_classes(mapping: dict[str, str]):
        return np.unique(list(mapping.values())).tolist()



from typing import TypeAlias

from .batch import BatchPredictionJob, BatchPredictionJobConfig
from .live import LivePredictionJob, LivePredictionJobConfig

PredictionJobConfig: TypeAlias = BatchPredictionJobConfig | LivePredictionJobConfig

__all__ = [
    "BatchPredictionJob",
    "LivePredictionJob",
    "PredictionJobConfig",
]



import regex


def is_storage_gs_uri(uri: str) -> bool:
    """
    Check if the given uri is a gs:// URI.

    Args:
        uri (str): The URI to check.

    Returns:
        bool: True if the URI is a valid Google Storage URI, False otherwise.
    """
    return regex.search(r"^gs://[a-z0-9][a-z0-9_-]{1,61}[a-z0-9]", uri) is not None


def is_storage_http_uri(uri) -> bool:
    """
    Check if the given uri is a storage.googleapis.com URI.

    Args:
        uri (str): The URI to check.

    Returns:
        bool: True if the URI is a valid Google Storage http URI, False otherwise.
    """
    return regex.search(r"^http[s]?://storage.googleapis.com", uri) is not None


def is_storage_appspot_uri(uri: str) -> bool:
    """
    Checks if the given uri is an appspot.com URI.
    Args:
        uri (str): The URI to check.

    Returns:
        bool: True if the URI is a valid appspot Google Storage URI, False otherwise.
    """
    return (
        regex.search(r"^http[s]?://.+.appspot.com.storage.googleapis.com", uri)
        is not None
    )


def get_gs_uri_from_uri(uri: str, appspot_name: str = "capturpwa") -> str:
    """Turns any Google Cloud Storage URL into a gs:// URI.

    Args:
        uri (str): The URI to convert.
        appspot_name (str, optional): The name of the appspot domain. Defaults to "capturpwa".

    Returns:
        str: The gs:// URI.

    Raises:
        ValueError: If the URL is not a valid GCS URL.
    """
    if is_storage_gs_uri(uri):
        return uri
    elif is_storage_http_uri(uri):
        return regex.sub(r"^http[s]?://storage.googleapis.com/*", "gs://", uri)
    elif is_storage_appspot_uri(uri):
        return regex.sub(
            r"^http[s]?://.+.appspot.com.storage.googleapis.com",
            f"gs://{appspot_name}.appspot.com",
            uri,
        )

    raise ValueError("Image URL is not a Google Cloud Storage URL.")


def get_bucket_from_gs_uri(gs_uri) -> str:
    """Extracts the Bucket name.

    Returns:
        str: Google Storage Bucket.
    """
    assert is_storage_gs_uri(gs_uri), "Not a valid Google Storage URI."

    exp = regex.search(r"^gs:\/\/([a-z0-9][a-z0-9_-]{1,61}[a-z0-9])", gs_uri)

    return None if exp is None else exp.group(1)


def get_filepath_from_gs_uri(gs_uri) -> str:
    """Extracts the filepath after the Bucket name.

    Returns:
        str: Filepath of the Google Storage URI.
    """
    assert is_storage_gs_uri(gs_uri), "Not a valid Google Storage URI."

    exp = regex.search(r"(?<=gs:\/\/[a-z0-9][a-z0-9_-]{1,61}[a-z0-9]\/).*", gs_uri)

    return None if exp is None else exp.group(0)



from typing import TypeAlias
from pydantic.dataclasses import dataclass as pd_dataclass

from captur_ml.core.dtypes.generic import ObjectWithId, WithConfidence
from captur_ml.core.dtypes.label import ClassLabel, BoundingBoxLabel


@pd_dataclass
class ClassPrediction(ClassLabel, WithConfidence):
    """A class prediction with a confidence score."""


@pd_dataclass
class BoundingBoxPrediction(BoundingBoxLabel, WithConfidence):
    """A bounding box prediction with a confidence score."""


Prediction: TypeAlias = ClassPrediction | BoundingBoxPrediction


@pd_dataclass
class PredictionSet(ObjectWithId):
    """A set of predictions."""

    #: A list of predictions
    predictions: list[Prediction]



from pydantic import root_validator
from pydantic.dataclasses import dataclass as pd_dataclass
from typing import Any, Literal

from .base import BaseStorageFolder
from captur_ml.core.exceptions import GoogleCloudStorageBucketNotFoundError
from captur_ml.core.utils.path import is_storage_gs_uri, get_bucket_from_gs_uri
from captur_ml.core.services.gcs import (
    _get_storage_client,
    write_file_to_gcs,
    write_json_to_gcs,
    write_jsonl_to_gcs,
    read_file_from_gcs,
    delete_file_from_gcs,
    check_file_exists,
    check_bucket_exists,
)


__pdoc__ = {
    "GCSFolder.bucket_exists": False,
}


@pd_dataclass
class GCSFolder(BaseStorageFolder):
    """A Google Cloud Storage folder."""

    #: The path to the folder in Google Cloud Storage.
    folder_path: str

    #: If true, run preflight checks on the folder path.
    check_exists: bool = True

    #: The client object used to interact with Google Cloud Storage.
    client = None

    def __post_init__(self):
        if self.client is None:
            self.client = _get_storage_client()

    @root_validator()
    def bucket_exists(cls, values):
        if values.get("check_exists") is False:
            return values

        full_path = values.get("folder_path")
        if not is_storage_gs_uri(full_path):
            raise ValueError(
                f"Expected a Google Cloud Storage URI, but received {full_path}"
            )

        bucket_name = get_bucket_from_gs_uri(full_path)
        if bucket_name is None or not check_bucket_exists(
            bucket_name, storage_client=values.get("client")
        ):
            raise GoogleCloudStorageBucketNotFoundError(
                f"The bucket {bucket_name} does not exist."
            )

        return values

    @property
    def bucket_name(self) -> str:
        """The name of the Google Cloud Storage bucket."""
        return get_bucket_from_gs_uri(self.folder_path)

    def write(self, file_contents: str | bytes, destination_filename: str):
        """Writes a file of either string or bytes content type to Google Cloud Storage.

        Args:
            file_contents (str | bytes): The contents of the file to write to Google Cloud Storage.
            destination_filename (str): The name of the file to write to Google Cloud Storage.

        Raises:
            GoogleCloudStorageResourceNotFoundError: If the Google Cloud Storage bucket does not exist.
        """
        if destination_filename.endswith(".json"):
            self._write_json(
                json_dict=file_contents,
                destination_filename=destination_filename,
            )
        elif destination_filename.endswith(".jsonl"):
            self._write_jsonl(
                json_lines=file_contents,
                bucket_name=self.bucket_name,
                destination_filename=destination_filename,
            )
        else:
            write_file_to_gcs(
                file_contents=file_contents,
                bucket_name=self.bucket_name,
                destination_filename=destination_filename,
                storage_client=self.client,
            )

    def _write_json(self, json_dict: dict[Any, Any], destination_filename: str):
        """Writes a JSON file to Google Cloud Storage.

        Args:
            json_dict (dict[Any, Any]): The dictionary to export as JSON to Google Cloud Storage.
            destination_filename (str): The name of the file to write to Google Cloud Storage. If it does not end with the .json extension, it will be added.

        Raises:
            ValueError: If the dictionary cannot be serialized to JSON.
        """
        write_json_to_gcs(
            json_dict=json_dict,
            bucket_name=self.bucket_name,
            destination_filename=destination_filename,
            storage_client=self.client,
        )

    def _write_jsonl(self, json_lines: list[dict[Any, Any]], destination_filename: str):
        """Writes a JSONL file to Google Cloud Storage.

        Args:
            json_lines (list[dict]): The list of dictionaries to export as JSONL to Google Cloud Storage.
            destination_filename (str): The name of the file to write to Google Cloud Storage. If it does not end with the .jsonl extension, it will be added.

        Raises:
            ValueError: If the dictionary cannot be serialized to JSON.
        """
        write_jsonl_to_gcs(
            json_lines=json_lines,
            bucket_name=self.bucket_name,
            destination_filename=destination_filename,
            storage_client=self.client,
        )

    def read(
        self, origin_filename: str, read_mode: Literal["r", "b"] = "r"
    ) -> bytes | str:
        """Reads a file from Google Cloud Storage.

        Args:
            origin_filename (str): The name of the file to read from Google Cloud Storage.
            read_mode (Literal["r", "b"], optional): The mode to read the file in. "r" for text, "b" for bytes. Defaults to "r".

        Raises:
            GoogleCloudStorageResourceNotFoundError: If the Google Cloud Storage bucket or file does not exist.
            ValueError: If the read_mode is not "b" but the file cannot be decoded as UTF-8.

        Returns:
            bytes | str: The raw contents of the file.
        """
        return read_file_from_gcs(
            bucket_name=self.bucket_name,
            origin_filename=origin_filename,
            read_mode=read_mode,
            storage_client=self.client,
        )

    def delete(self, destination_filename: str):
        """Deletes a file from Google Cloud Storage.

        Args:
            destination_filename (str): The name of the file to delete from Google Cloud Storage.

        Raises:
            GoogleCloudStorageResourceNotFoundError: If the Google Cloud Storage bucket or file does not exist.
        """
        delete_file_from_gcs(
            bucket_name=self.bucket_name,
            destination_filename=destination_filename,
            storage_client=self.client,
        )

    def contains(self, filename: str) -> bool:
        return check_file_exists(
            bucket_name=self.bucket_name,
            filename=filename,
            storage_client=self.client,
        )



from pydantic.dataclasses import dataclass as pd_dataclass

from captur_ml.core.dtypes import ImageOrImageList


@pd_dataclass
class BatchPredictionJobConfig:
    """
    Configuration for a batch prediction job.
    """

    #: The root directory in the cloud where the various files for the batch prediction job will be stored.
    root_directory: str


class BatchPredictionJob:
    """
    A batch prediction job is a job that runs on a batch of images and returns a promise type object that can be used to retrieve the result of the prediction job. It is first initialized with a configuration object and then started with a list of images.
    """

    def __init__(self, config: BatchPredictionJobConfig):
        self.config = config

    def start(self, images: ImageOrImageList):
        """Starts the (async) batch prediction job.

        Args:
            images (ImageOrImageList): The images to predict on.

        Raises:
            NotImplementedError
        """
        raise NotImplementedError()



from captur_ml.core.services.pubsub import publish
from captur_ml.core.dtypes import parsers


def parse_request(ml_gateway_request):
    """Responds to any HTTP request to MLGateway.
    Args:
        ml_gateway_request (dict) : the request made to the ML Gateway parsed using pydantic's parse_obj_as.
    Returns:
        Errors or Result (dict)
    """
    pubsub_topics = {
        "batch_predict": "ml-batch-prediction-pre-run",
        "live_predict": "ml-live-prediction",
        "model_train": "ml-train-pre-run",
        "model_evaluate": "ml-evaluation",
        "model_deploy": "ml-model-registration",
    }

    request_results = {
        "predict": [],
        "live_predict": [],
        "train": [],
        "evaluate": [],
        "register_model": [],
    }

    parser_mapping = {
        "batch_predict": parsers.parse_batch_predict_request,
        "live_predict": parsers.parse_live_predict_request,
        "model_train": parsers.parse_model_train_request,
        "model_evaluate": parsers.parse_mongo_model_evaluate_request,
        "deploy_model": parsers.parse_model_deploy_request,
    }

    for ml_operation, parser in parser_mapping.items():
        field = ml_gateway_request.dict()[ml_operation]
        if field:
            ml_operation_requests = parser(field)
            print(ml_operation_requests)
            for ml_operation_request in ml_operation_requests:
                publish(pubsub_topics[ml_operation], ml_operation_request)
                request_results[ml_operation].append(ml_operation_request)

    return {"result": request_results}



from dataclasses_json import dataclass_json, LetterCase
from pydantic.dataclasses import dataclass as pd_dataclass

from captur_ml.core.dtypes.generic import ObjectWithId
from captur_ml.core.dtypes.prediction import ClassPrediction


@dataclass_json(letter_case=LetterCase.CAMEL)
@pd_dataclass
class ImageClassificationResult(ObjectWithId):
    """The set of predictions for an image."""

    #: The set of predictions for the image.
    predictions: list[ClassPrediction]



from pydantic import AnyUrl, root_validator
from pydantic.dataclasses import dataclass as pd_dataclass

from captur_ml.core.dtypes.generic import ObjectWithId
from captur_ml.core.utils.image import load_image_bytes_from_url
from captur_ml.core.utils.path import get_gs_uri_from_uri


__pdoc__ = {
    "Image.must_contain_url_or_data": False,
    "Image.infer_mime": False,
}

MIME_TYPES = {
    "jpg": "image/jpeg",
    "jpeg": "image/jpeg",
    "png": "image/png",
    "tiff": "image/tiff",
    "tif": "image/tiff",
    "bmp": "image/bmp",
    "gif": "image/gif",
}


@pd_dataclass
class Image(ObjectWithId):
    """An image with optional metadata."""

    #: The image bytes encoded as a UTF-8 string. Required if `url` is not specified.
    data: str = ""
    #: An http[s]:// or gs:// URL pointing to the image resource. Required if `data` is not specified.
    url: AnyUrl | None = None
    #: The MIME type of the image. Required if `data` is specified.
    mime_type: str | None = None

    @root_validator
    def must_contain_url_or_data(cls, values):
        if not values.get("url") and not values.get("data"):
            raise ValueError("Either `url` or `data` must be specified.")
        return values

    @root_validator
    def infer_mime(cls, values):
        if values.get("mime_type") is not None:
            return values

        if values.get("url") is None:
            raise ValueError("If `url` is not used, `mime_type` must be specified.")

        suffix = values["url"].split(".")[-1]

        if suffix not in MIME_TYPES:
            raise ValueError(
                f"Unknown image mime type: {suffix}. Please specify `mime_type`."
            )

        values["mime_type"] = MIME_TYPES[suffix]

        return values

    @property
    def bytes(self):
        """The raw image bytes. Will make a network request if only `url` is specified and data has not been loaded yet."""
        if self.data:
            return self.data
        if self.url:
            self.data = load_image_bytes_from_url(self.url)
            return self.data
        raise ValueError("Cannot get bytes as image does not have a URL or data.")



from uuid import uuid4
from pydantic import confloat
from pydantic.dataclasses import dataclass as pd_dataclass


# The kw_only flag is set here to force the use of keyword arguments
# for the dataclass constructor. This is to avoid
#       TypeError: non-default argument '___' follows default argument


@pd_dataclass(kw_only=True)
class ObjectWithId:
    """This generic class automatically adds an id field to the dataclass.
    If the id field is not provided, it will be automatically generated as a UUID, using uuid4() from the Python uuid module.
    """

    #: A unique identifier of the object.
    id: str | None = None

    def __post_init__(self):
        if self.id is None:
            self.id = str(uuid4())


@pd_dataclass(kw_only=True)
class NamedObjectWithId(ObjectWithId):
    """This is a variant of the `captur_ml.core.dtypes.generic.ObjectWithId` class, which requires a name field."""

    #: A human-readable name of the object
    name: str


@pd_dataclass(kw_only=True)
class WithConfidence:
    """A generic object with a confidence score."""

    #: The statistical confidence of the Prediction. Must be between 0 and 1.
    confidence: confloat(ge=0, le=1)



from dataclasses import field
from pydantic.dataclasses import dataclass as pd_dataclass

from captur_ml.core.dtypes.generic import NamedObjectWithId, ObjectWithId


@pd_dataclass
class ClassLabel(NamedObjectWithId):
    """The classification label that describes the semantic meaning of an observation."""

    pass


@pd_dataclass
class AuditLabel(ObjectWithId):
    """A set of label that describes the acceptable classes of an observation during the audit process."""

    #: A list of human readable audit labels
    labels: list[ClassLabel] = field(default_factory=list)


@pd_dataclass
class BoundingBoxLabel(NamedObjectWithId):
    """The bounding box label that describes the semantic meaning of region in an observation."""

    #: The x-coordinate of the top-left corner of the bounding box
    x: float
    #: The y-coordinate of the top-left corner of the bounding box
    y: float
    #: The width of the bounding box
    width: float
    #: The height of the bounding box
    height: float



from abc import ABC
from pydantic.dataclasses import dataclass


@dataclass
class BaseStorageFolder(ABC):
    """A base class for storage folders. Storage folders are used to store images and other files."""

    folder_path: str

    def write(self, file_contents: str, destination_filename: str):
        pass

    def read(self, filename: str):
        pass

    def delete(self, filename: str):
        pass

    def contains(self, filename: str) -> bool:
        pass



from setuptools import setup, find_packages

long_desc = open("README.md").read()
required = [
    "pydantic",
    "google-cloud-storage",
    "google-cloud-aiplatform",
    "google-cloud-pubsub",
    "google-auth",
    "Pillow",
    "regex",
    "dataclasses-json",
    "numpy",
]

setup(
    name="captur-ml",
    version="0.3.1",
    description="The internal Captur Machine Learning SDK",
    long_description=long_desc,
    long_description_content_type="text/markdown",
    url="https://github.com/Captur/captur-ml-sdk",
    author=["Jack Barnett Leveson", "Jonny Jackson", "Shrinivasan Sankar"],
    author_email="jack@captur.ai",
    license="Creative Commons Attribution-NonCommercial-NoDerivatives 4.0 International",
    packages=find_packages(
        exclude=(
            "test",
            "docs",
        )
    ),
    install_requires=required,
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
    ],
    python_requires=">=3.10",
)



import re
from captur_ml.core.exceptions import (
    GoogleCloudVertexAIEndpointCorruptedImageError,
    GoogleCloudVertexAIEndpointDoesNotExistError,
    GoogleCloudVertexAIEndpointImageTooLargeError,
    GoogleCloudVertexAIEndpointNoModelDeployedError,
)
from captur_ml.core.entities import VertexAIEndpoint
from captur_ml.core.dtypes import Image, ImageOrImageList
from captur_ml.core.dtypes.result import ImageClassificationResult
from captur_ml.core.dtypes.prediction import ClassPrediction

from google.cloud import aiplatform_v1
from google.api_core.client_options import ClientOptions
from google.api_core import exceptions as google_exceptions
from google.cloud.aiplatform.gapic import PredictionServiceClient, schema


def get_image_classification_prediction_from_deployed_automl(
    images: ImageOrImageList,
    endpoint: VertexAIEndpoint,
    prediction_service_client_options: ClientOptions | None = None,
    run_preflight_checks: bool = False,
) -> aiplatform_v1.types.PredictResponse:
    """Gets image classification prediction from deployed automl model.

    Args:
        images (ImageOrImageList): The Image(s) to predict for.
        endpoint (VertexAIEndpoint): VertexAIEndpoint object instance that will be used to make the prediction.
        prediction_service_client_options (ClientOptions | None): Client options to pass to the prediction service client. Default = None.
        run_preflight_checks (bool): Whether to run preflight checks on the endpoint. Default = False.

    Raises:
        `captur_ml.core.exceptions.GoogleCloudVertexAIEndpointDoesNotExistError`: If the specified endpoint does not exist.
        `captur_ml.core.exceptions.GoogleCloudVertexAIEndpointNoModelDeployedError`: If there is no model deployed at the specified endpoint.
        `captur_ml.core.exceptions.GoogleCloudVertexAIEndpointImageTooLargeError`: If the image exceeds the 1.5MB limit.

    Returns:
        automl_v1beta1.types.PredictResponse: An object containing the prediction results for the image.
    """
    if run_preflight_checks:
        if not endpoint.exists:
            raise GoogleCloudVertexAIEndpointDoesNotExistError()
        if not endpoint.is_active:
            raise GoogleCloudVertexAIEndpointNoModelDeployedError()

    if prediction_service_client_options is None:
        prediction_service_client_options = {
            "api_endpoint": f"{endpoint.location}-aiplatform.googleapis.com"
        }

    client = PredictionServiceClient(client_options=prediction_service_client_options)

    if isinstance(images, Image):
        images = [images]

    instances = [
        schema.predict.instance.ImageClassificationPredictionInstance(
            content=image.bytes,
        ).to_value()
        for image in images
    ]

    parameters = schema.predict.params.ImageClassificationPredictionParams(
        confidence_threshold=0.0,
        max_predictions=10,
    ).to_value()

    endpoint = client.endpoint_path(
        project=endpoint.project,
        location=endpoint.location,
        endpoint=endpoint.id,
    )

    try:
        response = client.predict(
            endpoint=endpoint, instances=instances, parameters=parameters
        )
    except google_exceptions.FailedPrecondition as e:
        match = re.search(r"exceeds (\d*.*) limit", e.message)
        if match:
            raise GoogleCloudVertexAIEndpointImageTooLargeError()
        else:
            raise e
    except google_exceptions.InvalidArgument as e:
        raise GoogleCloudVertexAIEndpointCorruptedImageError()

    return response


def convert_automl_to_captur(
    images: ImageOrImageList, prediction_response: aiplatform_v1.types.PredictResponse
) -> list[ImageClassificationResult]:
    """Converts the automl_v1beta1.types.PredictResponse object to a captur_ml.core.dtypes.result.ImageClassificationResult object.

    Returns:
        `captur_ml.core.dtypes.result.ImageClassificationResult`: The converted object.
    """
    if isinstance(images, Image):
        images = [images]

    predictions = list(zip(images, prediction_response.predictions))

    results = []
    for image, prediction in predictions:
        ids = prediction["ids"]
        class_names = prediction["displayNames"]
        confidences = prediction["confidences"]

        predictions = []
        for id, name, confidence in zip(ids, class_names, confidences):
            pred = ClassPrediction(
                id=id,
                name=name,
                confidence=confidence,
            )
            predictions.append(pred)

        predictions.sort(key=lambda x: x.confidence, reverse=True)

        result = ImageClassificationResult(id=image.id, predictions=predictions)
        results.append(result)

    return results



import base64
import json
import pydantic

from flask.wrappers import Request


def parse_data_from_pubsub_message(
    event: bytes,
    pydantic_class: pydantic.main.ModelMetaclass,
):
    """Parses data from a Pub/Sub message and raises an error if the
       message has an invalid format.

    Args:
        event (bytes): Base64-encoded Pub/Sub message.
        pydantic_class (pydantic.main.ModelMetaclass): A specified pydantic class.

    Raises:
        ValueError: Raised if the message has an invalid format.

    Returns:
        PydanticClass: An instance of the specified pydantic class.
    """
    data = json.loads(base64.b64decode(event["data"]).decode("utf-8"))
    err = None
    try:
        event_data = pydantic.parse_obj_as(pydantic_class, data)
    except pydantic.ValidationError as e:
        err = e.errors()

    if err is not None:
        raise ValueError(f"Errors: {err}")

    return event_data


def parse_data_from_http_request(
    request: Request,
    pydantic_class: pydantic.main.ModelMetaclass,
):
    """Parses data from a HTTP Reuqest and raises an error if the
       message has an invalid format.

    Args:
        request (Request): Request sent to the Cloud Function.
        pydantic_class (pydantic.main.ModelMetaclass): A specified pydantic class.

    Raises:
        ValueError: Raised if the message has an invalid format.

    Returns:
        PydanticClass: An instance of the specified pydantic class.
    """
    request_json = request.get_json()
    err = None
    try:
        request_data = pydantic.parse_obj_as(pydantic_class, request_json)
    except pydantic.ValidationError as e:
        err = e.errors()

    if err is not None:
        raise ValueError(f"Errors: {err}")

    return request_data
